"""Events from pubsub subscriber."""

import datetime
import logging
from abc import ABC, abstractmethod

from .auth import AbstractAuth
from .registry import Registry
from .traits import BuildTraits, Command

EVENT_ID = "eventId"
EVENT_SESSION_ID = "eventSessionId"
TIMESTAMP = "timestamp"
RESOURCE_UPDATE = "resourceUpdate"
NAME = "name"
TRAITS = "traits"
EVENTS = "events"
RELATION_UPDATE = "relationUpdate"
TYPE = "type"
SUBJECT = "subject"
OBJECT = "object"

# Event images expire 30 seconds after the event is published
EVENT_IMAGE_EXPIRE_SECS = 30

EVENT_MAP = Registry()

_LOGGER = logging.getLogger(__name__)


class EventProcessingError(Exception):
    """Raised when there was an error handling an event."""


class ImageEventBase(ABC):
    """Base class for all image related event types."""

    def __init__(self, data: dict, timestamp: datetime.datetime):
        """Initialize EventBase."""
        self._data = data
        self._timestamp = timestamp

    @property
    def event_id(self) -> str:
        """An ID associated with the event.

        Can be used with CameraEventImageTrait to download the imaage.
        """
        return self._data[EVENT_ID]

    @property
    def event_session_id(self) -> str:
        """An ID used to associate separate messages with a single event."""
        return self._data[EVENT_SESSION_ID]

    @property
    def timestamp(self) -> datetime.datetime:
        """Timestap when the event occurred."""
        return self._timestamp

    @property
    def expires_at(self) -> datetime.datetime:
        return self._timestamp + datetime.timedelta(seconds=EVENT_IMAGE_EXPIRE_SECS)


@EVENT_MAP.register()
class CameraMotionEvent(ImageEventBase):
    """Motion has been detected by the camera."""

    NAME = "sdm.devices.events.CameraMotion.Motion"


@EVENT_MAP.register()
class CameraPersonEvent(ImageEventBase):
    """A person has been detected by the camera."""

    NAME = "sdm.devices.events.CameraPerson.Person"


@EVENT_MAP.register()
class CameraSoundEvent(ImageEventBase):
    """Sound has been detected by the camera."""

    NAME = "sdm.devices.events.CameraSound.Sound"


@EVENT_MAP.register()
class DoorbellChimeEvent(ImageEventBase):
    """The doorbell has been pressed."""

    NAME = "sdm.devices.events.DoorbellChime.Chime"


class RelationUpdate:
    """Represents a relational update for a resource."""

    def __init__(self, raw_data: dict):
        self._raw_data = raw_data

    @property
    def type(self) -> str:
        """The type of relation event 'CREATED', 'UPDATED', 'DELETED'."""
        return self._raw_data[TYPE]

    @property
    def subject(self) -> str:
        """The resource that the object is now in relation with."""
        return self._raw_data[SUBJECT]

    @property
    def object(self) -> str:
        """The resource that triggered the event."""
        return self._raw_data[OBJECT]


def BuildEvents(events: dict, event_map: dict, timestamp: datetime.datetime) -> dict:
    """Builds a trait map out of a response dict."""
    result = {}
    for (event, event_data) in events.items():
        if event not in event_map:
            continue
        cls = event_map[event]
        result[event] = cls(event_data, timestamp)
    return result


class EventMessage:
    """Event for a change in trait value or device action."""

    def __init__(self, raw_data: dict, auth: AbstractAuth):
        """Initialize an EventMessage."""
        _LOGGER.debug("EventMessage raw_data=%s", raw_data)
        self._raw_data = raw_data
        self._auth = auth

    @property
    def event_id(self) -> str:
        """A unique event identifier."""
        return self._raw_data.get(EVENT_ID, None)

    @property
    def timestamp(self) -> datetime.datetime:
        """Time when the event was published."""
        event_timestamp = self._raw_data[TIMESTAMP]
        return datetime.datetime.fromisoformat(event_timestamp.replace("Z", "+00:00"))

    @property
    def resource_update_name(self) -> str:
        """Returns the id of the device that was updated."""
        if RESOURCE_UPDATE not in self._raw_data:
            return None
        return self._raw_data[RESOURCE_UPDATE][NAME]

    @property
    def resource_update_events(self) -> dict:
        """Returns the set of events that happened."""
        if RESOURCE_UPDATE not in self._raw_data:
            return None
        events = self._raw_data[RESOURCE_UPDATE].get(EVENTS, {})
        return BuildEvents(events, EVENT_MAP, self.timestamp)

    @property
    def resource_update_traits(self) -> dict:
        """Returns the set of traits that were updated."""
        if RESOURCE_UPDATE not in self._raw_data:
            return None
        cmd = Command(self.resource_update_name, self._auth)
        events = self._raw_data[RESOURCE_UPDATE].get(TRAITS, {})
        return BuildTraits(events, cmd)

    @property
    def relation_update(self) -> RelationUpdate:
        """Represent a relational update for a resource."""
        if RELATION_UPDATE not in self._raw_data:
            return None
        return RelationUpdate(self._raw_data[RELATION_UPDATE])

    @property
    def raw_data(self) -> str:
        """Return the raw data string."""
        return self._raw_data


class AsyncEventCallback(ABC):
    """For implementers to get notified about EventMessages."""

    @abstractmethod
    async def async_handle_event(self, event_message: EventMessage):
        """Process an incoming EventMessage."""


class EventTypeFilterCallback(AsyncEventCallback):
    """Invokes a delegate only for events that match the trait type."""

    def __init__(self, event_name, delegate: AsyncEventCallback):
        self._event_name = event_name
        self._delegate = delegate

    async def async_handle_event(self, event_message: EventMessage):
        """Process an incoming EventMessage."""
        events = event_message.resource_update_events
        if self._event_name in events:
            await self._delegate.async_handle_event(event_message)


class RecentEventFilterCallback(AsyncEventCallback):
    """Invokes a delegate only for recent events."""

    def __init__(
        self, cutoff_timedelta: datetime.timedelta, delegate: AsyncEventCallback
    ):
        self._cutoff_timedelta = cutoff_timedelta
        self._delegate = delegate

    async def async_handle_event(self, event_message: EventMessage):
        """Process an incoming EventMessage."""
        if not event_message.timestamp:
            return
        now = datetime.datetime.now(tz=datetime.timezone.utc)
        oldest_allowed = now - self._cutoff_timedelta
        if event_message.timestamp < oldest_allowed:
            _LOGGER.debug(
                "Ignoring event that is too old (%s)", event_message.timestamp
            )
            return
        await self._delegate.async_handle_event(event_message)



"""Subscriber for the Smart Device Management event based API."""
import asyncio
import concurrent.futures
import json
import logging
import re
from abc import ABC, abstractmethod
from typing import Optional

from aiohttp.client_exceptions import ClientError
from google.api_core.exceptions import GoogleAPIError, NotFound, Unauthenticated
from google.cloud import pubsub_v1

from .auth import AbstractAuth
from .device_manager import DeviceManager
from .event import AsyncEventCallback, EventMessage
from .exceptions import AuthException, SubscriberException
from .google_nest_api import GoogleNestAPI

_LOGGER = logging.getLogger(__name__)

# Used to catch invalid subscriber id
EXPECTED_SUBSCRIBER_REGEXP = re.compile("projects/.*/subscriptions/.*")

# Used to catch a topic misconfiguration
EXPECTED_TOPIC_REGEXP = re.compile("projects/sdm-[a-z]+/topics/.*")


class AbstractSusbcriberFactory(ABC):
    """Abstract class for creating a subscriber, to facilitate testing."""

    @abstractmethod
    async def async_new_subscriber(
        self, creds, subscription_name, async_callback
    ) -> pubsub_v1.subscriber.futures.StreamingPullFuture:
        """Create a new event subscriber."""


class DefaultSubscriberFactory(AbstractSusbcriberFactory):
    """Default implementation that creates Google Pubsub subscriber."""

    async def async_new_subscriber(
        self, creds, subscription_name, loop, async_callback
    ) -> pubsub_v1.subscriber.futures.StreamingPullFuture:
        """Creates a new subscriber with a blocking to async bridge."""

        def callback_wrapper(message: pubsub_v1.subscriber.message.Message):
            future = asyncio.run_coroutine_threadsafe(async_callback(message), loop)
            future.result()

        with concurrent.futures.ThreadPoolExecutor(max_workers=1) as executor:
            return await loop.run_in_executor(
                executor,
                self._new_subscriber,
                creds,
                subscription_name,
                callback_wrapper,
            )

    def _new_subscriber(self, creds, subscription_name, callback_wrapper):
        """Issues a command to verify subscriber creds are correct."""
        subscriber = pubsub_v1.SubscriberClient(credentials=creds)
        subscription = subscriber.get_subscription(subscription=subscription_name)
        if subscription.topic:
            if not EXPECTED_TOPIC_REGEXP.match(subscription.topic):
                raise SubscriberException(
                    "Subscription misconfigured. Expected topic name to "
                    f"match '{EXPECTED_TOPIC_REGEXP.pattern}' but was "
                    f"'{subscription.topic}'."
                )
            else:
                _LOGGER.debug(
                    "Susbcriber '%s' configured on topic '%s'",
                    subscription_name,
                    subscription.topic,
                )
        return subscriber.subscribe(subscription_name, callback_wrapper)


class GoogleNestSubscriber:
    """Subscribes to events from the Google Nest feed."""

    def __init__(
        self,
        auth: AbstractAuth,
        project_id: str,
        subscriber_id: str,
        subscriber_factory: AbstractSusbcriberFactory = DefaultSubscriberFactory(),
        loop: Optional[asyncio.AbstractEventLoop] = None,
        watchdog_delay: float = 10,
    ):
        """Initialize the subscriber for the specified topic"""
        self._auth = auth
        self._subscriber_id = subscriber_id
        self._api = GoogleNestAPI(auth, project_id)
        self._loop = loop or asyncio.get_event_loop()
        self._subscriber_factory = subscriber_factory
        self._subscriber_future = None
        self._callback = None
        self._device_manager_task = asyncio.create_task(
            self._async_create_device_manager()
        )
        self._healthy = True
        self._watchdog_delay = watchdog_delay
        if self._watchdog_delay > 0:
            self._watchdog_task = asyncio.create_task(self._watchdog())
        else:
            self._watchdog_task = None

    def set_update_callback(self, callback: AsyncEventCallback):
        """Register a callback invoked when new messages are received."""
        self._callback = callback

    async def start_async(self):
        """Starts the subscriber."""
        if not EXPECTED_SUBSCRIBER_REGEXP.match(self._subscriber_id):
            raise SubscriberException(
                "Subscription misconfigured. Expected subscriber_id to "
                f"match '{EXPECTED_SUBSCRIBER_REGEXP.pattern}' but was "
                f"'{self._subscriber_id}'"
            )
        try:
            creds = await self._auth.async_get_creds()
        except ClientError as err:
            raise AuthException(f"Access token failure: {err}") from err

        try:
            self._subscriber_future = (
                await self._subscriber_factory.async_new_subscriber(
                    creds, self._subscriber_id, self._loop, self._async_message_callback
                )
            )
        except NotFound as err:
            raise SubscriberException(
                f"Failed to create subscriber '{self._subscriber_id}' id was not found"
            ) from err
        except Unauthenticated as err:
            raise AuthException("Failed to authenticate subscriber") from err
        except GoogleAPIError as err:
            raise SubscriberException(
                f"Failed to create subscriber '{self._subscriber_id}'"
            ) from err

        if not self._healthy:
            _LOGGER.info("Subscriber reconnected")
            self._healthy = True
        if self._subscriber_future:
            self._subscriber_future.add_done_callback(self._done_callback)

    async def _watchdog(self):
        """Background task that watches the subscriber and restarts it."""
        _LOGGER.debug("Starting background watchdog thread")
        while True:
            if self._subscriber_future and self._subscriber_future.done():
                _LOGGER.debug("Watchdog: subscriber shut down; restarting")
                await self.start_async()
            await asyncio.sleep(self._watchdog_delay)

    def wait(self):
        """Blocks on the subscriber."""
        self._subscriber_future.result()

    def stop_async(self):
        """Tells the subscriber to start shutting down."""
        if self._device_manager_task:
            self._device_manager_task.cancel()
        if self._watchdog_task:
            self._watchdog_task.cancel()
        if self._subscriber_future:
            self._subscriber_future.cancel()

    async def async_get_device_manager(self) -> DeviceManager:
        """Return the DeviceManger with the current state of devices."""
        return await self._device_manager_task

    async def _async_create_device_manager(self):
        """Creates a DeviceManager, populated with initial state."""
        device_manager = DeviceManager()
        structures = await self._api.async_get_structures()
        for structure in structures:
            device_manager.add_structure(structure)
        # Subscriber starts after a device fetch
        devices = await self._api.async_get_devices()
        for device in devices:
            device_manager.add_device(device)
        return device_manager

    def _done_callback(self, future):
        """Invoked when the subscriber is completed."""
        if future.exception():
            ex = future.exception()
            if self._healthy:
                self._healthy = False
                _LOGGER.warning(
                    "Subscriber disconnected, will restart: %s: %s", type(ex), ex
                )
            else:
                _LOGGER.debug("Subscriber failure: %s: %s", type(ex), ex)

    async def _async_message_callback(
        self, message: pubsub_v1.subscriber.message.Message
    ):
        """Invoked when a message is received."""
        payload = json.loads(bytes.decode(message.data))
        event = EventMessage(payload, self._auth)
        # Only accept device events once the Device Manager has been loaded.
        # We are ok with missing messages on startup since the device manager
        # will do a live read .
        if self._device_manager_task.done():
            await self._device_manager_task.result().async_handle_event(event)
        if self._callback:
            await self._callback.async_handle_event(event)
        message.ack()



import datetime

from .traits import TRAIT_MAP, Command

STATUS = "status"
TIMER_MODE = "timerMode"
TIMER_TIMEOUT = "timerTimeout"
CUSTOM_NAME = "customName"
AMBIENT_HUMIDITY_PERCENT = "ambientHumidityPercent"
AMBIENT_TEMPERATURE_CELSIUS = "ambientTemperatureCelsius"


@TRAIT_MAP.register()
class ConnectivityTrait:
    """This trait belongs to any device that has connectivity information."""

    NAME = "sdm.devices.traits.Connectivity"

    def __init__(self, data: dict, cmd: Command):
        self._data = data

    @property
    def status(self) -> str:
        """Device connectivity status.

        Return:
          "OFFLINE", "ONLINE"
        """
        return self._data[STATUS]


@TRAIT_MAP.register()
class FanTrait:
    """This trait belongs to any device that can control the fan."""

    NAME = "sdm.devices.traits.Fan"

    def __init__(self, data: dict, cmd: Command):
        self._data = data
        self._cmd = cmd

    @property
    def timer_mode(self) -> str:
        """Current timer mode.

        Return:
          "ON", "OFF"
        """
        return self._data.get(TIMER_MODE)

    @property
    def timer_timeout(self) -> datetime.datetime:
        """Timestamp at which timer mode will turn to OFF."""
        if TIMER_TIMEOUT not in self._data:
            return None
        timestamp = self._data[TIMER_TIMEOUT]
        return datetime.datetime.fromisoformat(timestamp.replace("Z", "+00:00"))

    async def set_timer(self, timer_mode, duration=None):
        """Change the fan timer."""
        data = {
            "command": "sdm.devices.commands.Fan.SetTimer",
            "params": {
                "timerMode": timer_mode,
            },
        }
        if duration:
            data["params"]["duration"] = f"{duration}s"
        return await self._cmd.execute(data)


@TRAIT_MAP.register()
class InfoTrait:
    """This trait belongs to any device for device-related information."""

    NAME = "sdm.devices.traits.Info"

    def __init__(self, data: dict, cmd: Command):
        self._data = data

    @property
    def custom_name(self) -> str:
        """Custom name of the device."""
        return self._data[CUSTOM_NAME]


@TRAIT_MAP.register()
class HumidityTrait:
    """This trait belongs to any device that has a sensor to measure humidity."""

    NAME = "sdm.devices.traits.Humidity"

    def __init__(self, data: dict, cmd: Command):
        self._data = data

    @property
    def ambient_humidity_percent(self) -> float:
        """Percent humidity, measured at the device."""
        return self._data[AMBIENT_HUMIDITY_PERCENT]


@TRAIT_MAP.register()
class TemperatureTrait:
    """This trait belongs to any device that has a sensor to measure temperature."""

    NAME = "sdm.devices.traits.Temperature"

    def __init__(self, data: dict, cmd: Command):
        self._data = data

    @property
    def ambient_temperature_celsius(self) -> float:
        """Percent humidity, measured at the device."""
        return self._data[AMBIENT_TEMPERATURE_CELSIUS]



"""Library for exceptions using the Google Nest SDM API and subscriber."""


class GoogleNestException(Exception):
    """Base class for all client exceptions."""


class SubscriberException(GoogleNestException):
    """Raised during problems subscribing to events and updates."""


class ApiException(GoogleNestException):
    """Raised during problems talking to the API."""


class AuthException(GoogleNestException):
    """Raised due to auth problems talking to API or subscriber."""



"""Device Manager keeps track of the current state of all devices."""

from .device import Device
from .event import AsyncEventCallback, EventMessage
from .structure import InfoTrait, RoomInfoTrait, Structure


class DeviceManager(AsyncEventCallback):
    """DeviceManager holds current state of all devices."""

    def __init__(self):
        self._devices = {}
        self._structures = {}
        self._callback = None

    @property
    def devices(self) -> dict:
        """Return current state of devices."""
        return self._devices

    @property
    def structures(self) -> dict:
        """Return current state of structures."""
        return self._structures

    def add_device(self, device: Device):
        """Tracks the specified device."""
        self._devices[device.name] = device

    def add_structure(self, structure: Structure):
        """Tracks the specified device."""
        self._structures[structure.name] = structure

    async def async_handle_event(self, event_message: EventMessage):
        """Invokes by the subscriber when a new message is received."""
        if event_message.relation_update:
            relation = event_message.relation_update
            if relation.object in self._devices:
                device = self._devices[relation.object]
                if relation.type == "DELETED":
                    # Delete device from room/structure
                    if relation.subject in device.parent_relations:
                        del device.parent_relations[relation.subject]

                # Device moved to a room
                if relation.type == "UPDATED" or relation.type == "CREATED":
                    name = "Unknown"
                    if relation.subject in self._structures:
                        structure = self._structures[relation.subject]
                        for trait_name in [InfoTrait.NAME, RoomInfoTrait.NAME]:
                            if trait_name in structure.traits:
                                name = structure.traits[trait_name].custom_name
                                continue
                    device.parent_relations[relation.subject] = name
        if event_message.resource_update_name:
            device_id = event_message.resource_update_name
            if device_id in self._devices:
                device = self._devices[device_id]
                await device.async_handle_event(event_message)



"""Authentication library, implemented by users of the API."""
import logging
from abc import ABC, abstractmethod

import aiohttp
from aiohttp.client_exceptions import ClientError, ClientResponseError
from google.auth.credentials import Credentials
from google.oauth2.credentials import Credentials as OAuthCredentials

from .exceptions import ApiException, AuthException

HTTP_UNAUTHORIZED = 401


class AbstractAuth(ABC):
    """Abstract class to make authenticated requests."""

    def __init__(self, websession: aiohttp.ClientSession, host: str):
        """Initialize the auth."""
        self._websession = websession
        self._host = host

    @abstractmethod
    async def async_get_access_token(self) -> str:
        """Return a valid access token."""

    async def async_get_creds(self) -> Credentials:
        """Return creds for subscriber API."""
        token = await self.async_get_access_token()
        return OAuthCredentials(token=token)

    async def request(self, method: str, url: str, **kwargs) -> aiohttp.ClientResponse:
        """Make a request."""
        headers = kwargs.get("headers")

        if headers is None:
            headers = {}
        else:
            headers = dict(headers)
            del kwargs["headers"]

        try:
            access_token = await self.async_get_access_token()
        except ClientError as err:
            raise AuthException(f"Access token failure: {err}") from err
        headers["authorization"] = f"Bearer {access_token}"
        url = f"{self._host}/{url}"
        logging.debug("request[%s]=%s", method, url)
        return await self._websession.request(method, url, **kwargs, headers=headers)

    @staticmethod
    def raise_for_status(resp: aiohttp.ClientResponse) -> aiohttp.ClientResponse:
        """Raise exceptions on failure methods."""
        try:
            resp.raise_for_status()
        except ClientResponseError as err:
            if err.status == HTTP_UNAUTHORIZED:
                raise AuthException(f"Unable to authenticate with API: {err}") from err
            raise ApiException(f"Error from API: {err}") from err
        except ClientError as err:
            raise ApiException(f"Error from API: {err}") from err
        return resp



"""Traits belonging to doorbell devices."""

from .traits import TRAIT_MAP, Command


@TRAIT_MAP.register()
class DoorbellChimeTrait:
    """For any device that supports a doorbell chime and related press events."""

    NAME = "sdm.devices.traits.DoorbellChime"

    def __init__(self, data: dict, cmd: Command):
        """Initialize DoorbellChime."""
        self._data = data
        self._cmd = cmd



"""Traits for thermostats."""
from .traits import TRAIT_MAP, Command

STATUS = "status"
AVAILABLE_MODES = "availableModes"
MODE = "mode"
HEAT_CELSIUS = "heatCelsius"
COOL_CELSIUS = "coolCelsius"


@TRAIT_MAP.register()
class ThermostatEcoTrait:
    """This trait belongs to any device that has a sensor to measure temperature."""

    NAME = "sdm.devices.traits.ThermostatEco"

    def __init__(self, data: dict, cmd: Command):
        """Initialize ThermostatEcoTrait."""
        self._data = data
        self._cmd = cmd

    @property
    def available_modes(self) -> list:
        """List of supported Eco modes."""
        return self._data[AVAILABLE_MODES]

    @property
    def mode(self) -> str:
        """The current Eco mode of the thermostat."""
        return self._data[MODE]

    async def set_mode(self, mode):
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatEco.SetMode",
            "params": {"mode": mode},
        }
        return await self._cmd.execute(data)

    @property
    def heat_celsius(self) -> float:
        """Lowest temperature where Eco mode begins heating."""
        return self._data[HEAT_CELSIUS]

    @property
    def cool_celsius(self) -> float:
        """Highest cooling temperature where Eco mode begins cooling."""
        return self._data[COOL_CELSIUS]


@TRAIT_MAP.register()
class ThermostatHvacTrait:
    """This trait belongs to devices that can report HVAC details."""

    NAME = "sdm.devices.traits.ThermostatHvac"

    def __init__(self, data: dict, cmd: Command):
        """Initialize ThermostatHvacTrait."""
        self._data = data

    @property
    def status(self) -> list:
        """Current HVAC status of the thermostat."""
        return self._data[STATUS]


@TRAIT_MAP.register()
class ThermostatModeTrait:
    """This trait belongs to devices that support different thermostat modes."""

    NAME = "sdm.devices.traits.ThermostatMode"

    def __init__(self, data: dict, cmd: Command):
        """Initialize ThermostatModeTrait."""
        self._data = data
        self._cmd = cmd

    @property
    def available_modes(self) -> list:
        """List of supported thermostat modes."""
        return self._data[AVAILABLE_MODES]

    @property
    def mode(self) -> str:
        """The current mode of the thermostat."""
        return self._data[MODE]

    async def set_mode(self, mode):
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatMode.SetMode",
            "params": {"mode": mode},
        }
        return await self._cmd.execute(data)


@TRAIT_MAP.register()
class ThermostatTemperatureSetpointTrait:
    """This trait belongs to devices that support setting target temperature."""

    NAME = "sdm.devices.traits.ThermostatTemperatureSetpoint"

    def __init__(self, data: dict, cmd: Command):
        """Initialize ThermostatTemperatureSetpointTrait."""
        self._data = data
        self._cmd = cmd

    @property
    def heat_celsius(self) -> float:
        """Lowest temperature where Eco mode begins heating."""
        return self._data[HEAT_CELSIUS]

    @property
    def cool_celsius(self) -> list:
        """Highest cooling temperature where Eco mode begins cooling."""
        return self._data[COOL_CELSIUS]

    async def set_heat(self, heat: float):
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatTemperatureSetpoint.SetHeat",
            "params": {"heatCelsius": heat},
        }
        return await self._cmd.execute(data)

    async def set_cool(self, cool: float):
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatTemperatureSetpoint.SetCool",
            "params": {"coolCelsius": cool},
        }
        return await self._cmd.execute(data)

    async def set_range(self, heat: float, cool: float):
        """Change the thermostat Eco mode."""
        data = {
            "command": "sdm.devices.commands.ThermostatTemperatureSetpoint.SetRange",
            "params": {
                "heatCelsius": heat,
                "coolCelsius": cool,
            },
        }
        return await self._cmd.execute(data)



"""Traits for structures / rooms."""

STRUCTURE_NAME = "name"
STRUCTURE_TRAITS = "traits"
CUSTOM_NAME = "customName"


class InfoTrait:
    """This trait belongs to any structure for structure-related information."""

    NAME = "sdm.structures.traits.Info"

    def __init__(self, data: dict):
        self._data = data

    @property
    def custom_name(self) -> str:
        """Custom name of the structure."""
        return self._data[CUSTOM_NAME]


class RoomInfoTrait:
    """This trait belongs to any structure for room-related information."""

    NAME = "sdm.structures.traits.RoomInfo"

    def __init__(self, data: dict):
        self._data = data

    @property
    def custom_name(self) -> str:
        """Custom name of the room."""
        return self._data[CUSTOM_NAME]


_ALL_TRAITS = [
    InfoTrait,
    RoomInfoTrait,
]
_ALL_TRAIT_MAP = {cls.NAME: cls for cls in _ALL_TRAITS}


def _TraitsDict(traits: dict, trait_map: dict):
    result = {}
    for (trait, trait_data) in traits.items():
        if trait not in trait_map:
            continue
        cls = trait_map[trait]
        result[trait] = cls(trait_data)
    return result


class Structure:
    """Class that represents a structure object in the Google Nest SDM API."""

    def __init__(self, raw_data: dict, traits: dict):
        """Initialize a structure."""
        self._raw_data = raw_data
        self._traits = traits

    @staticmethod
    def MakeStructure(raw_data: dict):
        """Creates a structure with the appropriate traits."""
        traits = raw_data.get(STRUCTURE_TRAITS, {})
        traits_dict = _TraitsDict(traits, _ALL_TRAIT_MAP)
        return Structure(raw_data, traits_dict)

    @property
    def name(self) -> str:
        """The resource name of the structure e.g. 'enterprises/XYZ/structures/123'."""
        return self._raw_data[STRUCTURE_NAME]

    @property
    def traits(self) -> dict:
        """Return a trait mixin on None."""
        return self._traits

    def _traits_data(self, trait) -> dict:
        """Return the raw dictionary for the specified trait."""
        traits_dict = self._raw_data.get(STRUCTURE_TRAITS, {})
        return traits_dict.get(trait, {})

    @property
    def raw_data(self) -> str:
        """Return the raw data string."""
        return self._raw_data



"""Traits belonging to camera devices."""

import datetime
import urllib.parse as urlparse

from .traits import TRAIT_MAP, Command

MAX_IMAGE_RESOLUTION = "maxImageResolution"
MAX_VIDEO_RESOLUTION = "maxVideoResolution"
WIDTH = "width"
HEIGHT = "height"
VIDEO_CODECS = "videoCodecs"
AUDIO_CODECS = "audioCodecs"
STREAM_URLS = "streamUrls"
RESULTS = "results"
RTSP_URL = "rtspUrl"
STREAM_EXTENSION_TOKEN = "streamExtensionToken"
STREAM_TOKEN = "streamToken"
URL = "url"
TOKEN = "token"
EXPIRES_AT = "expiresAt"


class Resolution:
    """Maximum Resolution of an image or stream."""

    width = None
    height = None


@TRAIT_MAP.register()
class CameraImageTrait:
    """This trait belongs to any device that supports taking images."""

    NAME = "sdm.devices.traits.CameraImage"

    def __init__(self, data: dict, cmd: Command):
        """Initialize CameraImageTrait."""
        self._data = data
        self._cmd = cmd

    @property
    def max_image_resolution(self) -> Resolution:
        """Maximum resolution of the camera image."""
        res = Resolution()
        res.width = self._data[MAX_IMAGE_RESOLUTION][WIDTH]
        res.height = self._data[MAX_IMAGE_RESOLUTION][HEIGHT]
        return res


class RtspStream:
    """Provides access an RTSP live stream URL."""

    def __init__(self, data: dict, cmd: Command):
        """Initialize RstpStream."""
        self._data = data
        self._cmd = cmd

    @property
    def rtsp_stream_url(self) -> str:
        """RTSP live stream URL."""
        return self._data[STREAM_URLS][RTSP_URL]

    @property
    def stream_token(self) -> str:
        """Token to use to access an RTSP live stream."""
        return self._data[STREAM_TOKEN]

    @property
    def stream_extension_token(self) -> str:
        """Token to use to access an RTSP live stream."""
        return self._data[STREAM_EXTENSION_TOKEN]

    @property
    def expires_at(self) -> datetime.datetime:
        """Time at which both streamExtensionToken and streamToken expire."""
        expires_at = self._data[EXPIRES_AT]
        return datetime.datetime.fromisoformat(expires_at.replace("Z", "+00:00"))

    async def extend_rtsp_stream(self):
        """Request a new RTSP live stream URL access token."""
        data = {
            "command": "sdm.devices.commands.CameraLiveStream.ExtendRtspStream",
            "params": {"streamExtensionToken": self.stream_extension_token},
        }
        resp = await self._cmd.execute(data)
        resp.raise_for_status()
        response_data = await resp.json()
        results = response_data[RESULTS]
        # Update the stream url with the new token
        stream_token = results[STREAM_TOKEN]
        parsed = urlparse.urlparse(self.rtsp_stream_url)
        parsed = parsed._replace(query=f"auth={stream_token}")
        url = urlparse.urlunparse(parsed)
        results[STREAM_URLS] = {}
        results[STREAM_URLS][RTSP_URL] = url
        return RtspStream(results, self._cmd)

    async def stop_rtsp_stream(self):
        """Invalidates a valid RTSP access token and stops the RTSP live stream."""
        data = {
            "command": "sdm.devices.commands.CameraLiveStream.StopRtspStream",
            "params": {"streamExtensionToken": self.stream_extension_token},
        }
        resp = await self._cmd.execute(data)
        resp.raise_for_status()


@TRAIT_MAP.register()
class CameraLiveStreamTrait:
    """This trait belongs to any device that supports live streaming."""

    NAME = "sdm.devices.traits.CameraLiveStream"

    def __init__(self, data: dict, cmd: Command):
        """Initialize CameraLiveStreamTrait."""
        self._data = data
        self._cmd = cmd

    @property
    def max_video_resolution(self) -> Resolution:
        """Maximum resolution of the video live stream."""
        res = Resolution()
        res.width = self._data[MAX_VIDEO_RESOLUTION][WIDTH]
        res.height = self._data[MAX_VIDEO_RESOLUTION][HEIGHT]
        return res

    @property
    def video_codecs(self) -> list:
        """Video codecs supported for the live stream."""
        return self._data[VIDEO_CODECS]

    @property
    def audio_codecs(self) -> list:
        """Audio codecs supported for the live stream."""
        return self._data[AUDIO_CODECS]

    async def generate_rtsp_stream(self) -> RtspStream:
        """Request a token to access an RTSP live stream URL."""
        data = {
            "command": "sdm.devices.commands.CameraLiveStream.GenerateRtspStream",
            "params": {},
        }
        resp = await self._cmd.execute(data)
        resp.raise_for_status()
        response_data = await resp.json()
        results = response_data[RESULTS]
        return RtspStream(results, self._cmd)


class EventImage:
    """Provides access to an image in response to an event.

    Use a ?width or ?height query parameters to customize the resolution
    of the downloaded image. Only one of these parameters need to specified.
    The other parameter is scaled automatically according to the camera's
    aspect ratio.

    The token should be added as an HTTP header:
    Authorization: Basic <token>
    """

    def __init__(self, data: dict, cmd: Command):
        self._data = data
        self._cmd = cmd

    @property
    def url(self) -> str:
        """The URL to download the camera image from."""

        return self._data[URL]

    @property
    def token(self) -> str:
        """Token to use in the HTTP Authorization header when downloading."""
        return self._data[TOKEN]


@TRAIT_MAP.register()
class CameraEventImageTrait:
    """This trait belongs to any device that generates images from events."""

    NAME = "sdm.devices.traits.CameraEventImage"

    def __init__(self, data: dict, cmd: Command):
        """Initialize CameraEventImageTrait."""
        self._data = data
        self._cmd = cmd

    async def generate_image(self, event_id: str) -> EventImage:
        """Provides a URL to download a camera image from."""
        data = {
            "command": "sdm.devices.commands.CameraEventImage.GenerateImage",
            "params": {
                "eventId": event_id,
            },
        }
        resp = await self._cmd.execute(data)
        resp.raise_for_status()
        response_data = await resp.json()
        results = response_data[RESULTS]
        return EventImage(results, self._cmd)


@TRAIT_MAP.register()
class CameraMotionTrait:
    """For any device that supports motion detection events."""

    NAME = "sdm.devices.traits.CameraMotion"

    def __init__(self, data: dict, cmd: Command):
        """Initialize CameraMotionTrait."""
        self._data = data
        self._cmd = cmd


@TRAIT_MAP.register()
class CameraPersonTrait:
    """For any device that supports person detection events."""

    NAME = "sdm.devices.traits.CameraPerson"

    def __init__(self, data: dict, cmd: Command):
        """Initialize CameraPersonTrait."""
        self._data = data
        self._cmd = cmd


@TRAIT_MAP.register()
class CameraSoundTrait:
    """For any device that supports sound detection events."""

    NAME = "sdm.devices.traits.CameraSound"

    def __init__(self, data: dict, cmd: Command):
        """Initialize CameraSoundTrait."""
        self._data = data
        self._cmd = cmd



import aiohttp

from .auth import AbstractAuth
from .registry import Registry

DEVICE_TRAITS = "traits"

TRAIT_MAP = Registry()


class Command:
    """Base class for executing commands."""

    def __init__(self, device_id: str, auth: AbstractAuth):
        self._device_id = device_id
        self._auth = auth

    async def execute(self, data: dict) -> aiohttp.ClientResponse:
        resp = await self._auth.request(
            "post", f"{self._device_id}:executeCommand", json=data
        )
        return AbstractAuth.raise_for_status(resp)


def _TraitsDict(traits: dict, trait_map: dict, cmd: Command):
    d = {}
    for (trait, trait_data) in traits.items():
        if trait not in trait_map:
            continue
        cls = trait_map[trait]
        d[trait] = cls(trait_data, cmd)
    return d


def BuildTraits(traits: dict, cmd: Command, device_type=None) -> dict:
    """Builds a trait map out of a response dict."""
    # There is a bug where doorbells do not return the DoorbellChime trait.  Simulate
    # that it was returned
    if device_type and device_type == "sdm.devices.types.DOORBELL":
        traits = traits.copy()
        traits["sdm.devices.traits.DoorbellChime"] = {}
    return _TraitsDict(traits, TRAIT_MAP, cmd)



import pathlib
from setuptools import setup

VERSION = '0.2.1'

HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()

setup(name='google_nest_sdm',
      version=VERSION,
      description='Library for the Google Nest SDM API',
      long_description=README,
      long_description_content_type="text/markdown",
      keywords='google nest sdm camera therostat security doorbell',
      author='Allen Porter',
      author_email='allen@thebends.org',
      url='https://github.com/allenporter/python-google-nest-sdm',
      packages=['google_nest_sdm'],
      include_package_data=True,
      install_requires=[
          'aiohttp>=3.6.2',
          'google-auth>=1.22.0',
          'google-auth-oauthlib>=0.4.1',
          'google-cloud-pubsub>=2.1.0',
          'requests-oauthlib>=1.3.0',
      ],
      entry_points = {
          'console_scripts': [
              'google_nest=google_nest_sdm.google_nest:main',
          ],
      }
)



#!/usr/bin/python3

"""Command line tool for the Google Nest Smart Device Management API

You must configure your device as described:
  https://developers.google.com/nest/device-access/get-started

which will give you a project_id, client_id, and client_secret.  This tool
will do a one time setup to get an access token, then from there on will
cache the token on local disk.

Once authenticated, you can run commands like:

$ google_nest --project_id=<project_id> list
$ google_nest --project_id=<project_id> get <device_id>
"""

import argparse
import asyncio
import errno
import json
import logging
import os
import pickle
import socket

import yaml
from aiohttp import ClientSession, TCPConnector
from google.auth.credentials import Credentials
from google.auth.transport.requests import Request
from google.oauth2 import service_account
from google_auth_oauthlib.flow import InstalledAppFlow

from .auth import AbstractAuth
from .camera_traits import CameraLiveStreamTrait
from .event import EventMessage
from .google_nest_api import GoogleNestAPI
from .google_nest_subscriber import AsyncEventCallback, GoogleNestSubscriber
from .thermostat_traits import (
    ThermostatEcoTrait,
    ThermostatModeTrait,
    ThermostatTemperatureSetpointTrait,
)

# Define command line arguments
parser = argparse.ArgumentParser(
    description="Command line tool for Google Nest SDM API"
)
parser.add_argument("--project_id", required=True, help="Device Access program id")
parser.add_argument("--creds_json_file", help="Service acccount credentials file")
parser.add_argument("--client_id", help="OAuth credentials client_id")
parser.add_argument("--client_secret", help="OAuth credentials client_secret")
parser.add_argument(
    "--token_cache",
    help="File storage for long lived creds",
    default="~/.config/google_nest/token_cache",
)
parser.add_argument(
    "-v", "--verbose", help="Increase output verbosity", action="store_true"
)
parser.add_argument(
    "--output_type",
    type=str,
    choices=["json", "yaml"],
    help="Change the output type from json or yaml (default).",
    default="yaml",
)

cmd_parser = parser.add_subparsers(dest="command", required=True)
list_structures_parser = cmd_parser.add_parser("list_structures")
list_devices_parser = cmd_parser.add_parser("list_devices")
get_structure_parser = cmd_parser.add_parser("get_structure")
get_structure_parser.add_argument("structure_id")
get_device_parser = cmd_parser.add_parser("get_device")
get_device_parser.add_argument("device_id")
set_mode_parser = cmd_parser.add_parser(
    "set_mode", description="Change the thermostat mode."
)
set_mode_parser.add_argument("device_id")
set_mode_parser.add_argument(
    "mode",
    help="The mode to change the thermostat to.",
    choices=["MANUAL_ECO", "HEAT", "COOL", "HEATCOOL", "OFF"],
)
set_heat_parser = cmd_parser.add_parser(
    "set_heat", description="Sets the target temperature when in HEAT mode."
)
set_heat_parser.add_argument("device_id")
set_heat_parser.add_argument("heat", type=float)
set_cool_parser = cmd_parser.add_parser(
    "set_cool", help="Sets the target temperature when in COOL mode."
)
set_cool_parser.add_argument("device_id")
set_cool_parser.add_argument(
    "cool",
    type=float,
    help="The target temperature to set when the thermostat is in COOL mode.",
)
set_range_parser = cmd_parser.add_parser(
    "set_range", help="Sets the min/max temperature when in HEATCOOL mode."
)
set_range_parser.add_argument("device_id")
set_range_parser.add_argument(
    "heat", type=float, help="The minimum target temperature to set."
)
set_range_parser.add_argument(
    "cool", type=float, help="The maximum target temperature to set."
)
generate_rtsp_stream_parser = cmd_parser.add_parser("generate_rtsp_stream")
generate_rtsp_stream_parser.add_argument("device_id")
subscribe_parser = cmd_parser.add_parser("subscribe")
subscribe_parser.add_argument("subscription_id")
subscribe_parser.add_argument("device_id", nargs="?")

OAUTH2_AUTHORIZE = (
    "https://nestservices.google.com/partnerconnections/{project_id}/auth"
)
OAUTH2_TOKEN = "https://www.googleapis.com/oauth2/v4/token"
SDM_SCOPES = [
    "https://www.googleapis.com/auth/sdm.service",
    "https://www.googleapis.com/auth/pubsub",
]
API_URL = "https://smartdevicemanagement.googleapis.com/v1"


class Auth(AbstractAuth):
    """Implementation of AbstractAuth that uses the token cache."""

    def __init__(
        self,
        websession: ClientSession,
        user_creds: Credentials,
        service_creds: Credentials,
        api_url: str,
    ):
        """Initialize Google Nest Device Access auth."""
        super().__init__(websession, api_url)
        self._user_creds = user_creds
        self._service_creds = service_creds

    async def async_get_access_token(self):
        """Return a valid access token."""
        return self._user_creds.token

    async def async_get_creds(self):
        return self._service_creds


def CreateCreds(args) -> Credentials:
    """Runs an interactive flow to get OAuth creds."""

    creds = None
    token_cache = os.path.expanduser(args.token_cache)
    if os.path.exists(token_cache):
        with open(token_cache, "rb") as token:
            creds = pickle.load(token)

    # If there are no (valid) credentials available, let the user log in.
    if not creds or not creds.valid:
        if creds and creds.expired and creds.refresh_token:
            creds.refresh(Request())
        else:
            if not args.client_id or not args.client_secret:
                raise ValueError("Required flag --client_id or --client_secret missing")
            client_config = {
                "installed": {
                    "client_id": args.client_id,
                    "client_secret": args.client_secret,
                    "redirect_uris": ["urn:ietf:wg:oauth:2.0:oob"],
                    "auth_uri": OAUTH2_AUTHORIZE.format(project_id=args.project_id),
                    "token_uri": OAUTH2_TOKEN,
                },
            }
            app_flow = InstalledAppFlow.from_client_config(
                client_config, scopes=SDM_SCOPES
            )
            creds = app_flow.run_console()
        # Save the credentials for the next run
        if not os.path.exists(os.path.dirname(token_cache)):
            try:
                os.makedirs(os.path.dirname(token_cache))
            except OSError as exc:  # Guard against race condition
                if exc.errno != errno.EEXIST:
                    raise
        with open(token_cache, "wb") as token:
            pickle.dump(creds, token)
    return creds


def PrintStructure(structure, output_type):
    if output_type == "json":
        print(json.dumps(structure.raw_data))
    else:
        print(yaml.dump(structure.raw_data))


def PrintDevice(device, output_type):
    if output_type == "json":
        print(json.dumps(device.raw_data))
    else:
        print(yaml.dump(device.raw_data))


class SubscribeCallback(AsyncEventCallback):
    def __init__(self, output_type=None):
        self._output_type = output_type

    async def async_handle_event(self, event_message: EventMessage):
        if self._output_type == "json":
            print(json.dumps(event_message.raw_data))
        else:
            print(yaml.dump(event_message.raw_data))


class DeviceWatcherCallback(AsyncEventCallback):
    def __init__(self, device, output_type):
        self._device = device
        self._output_type = output_type

    async def async_handle_event(self, event_message: EventMessage):
        print(f"event_id: {event_message.event_id}")
        print("Current device state:")
        PrintDevice(self._device, self._output_type)
        print("")


async def RunTool(args, user_creds: Credentials, service_creds: Credentials):
    conn = TCPConnector(
        family=socket.AF_INET,
        verify_ssl=False,
    )
    async with ClientSession(connector=conn) as client:
        auth = Auth(client, user_creds, service_creds, API_URL)
        api = GoogleNestAPI(auth, args.project_id)

        if args.command == "list_structures":
            structures = await api.async_get_structures()
            for structure in structures:
                PrintStructure(structure, args.output_type)
            return

        if args.command == "get_structure":
            structure = await api.async_get_structure(args.structure_id)
            PrintStructure(structure, args.output_type)
            return

        if args.command == "list_devices":
            devices = await api.async_get_devices()
            for device in devices:
                PrintDevice(device, args.output_type)
            return

        if args.command == "subscribe":
            logging.info("Subscription: %s", args.subscription_id)
            subscriber = GoogleNestSubscriber(
                auth, args.project_id, args.subscription_id
            )
            if args.device_id:
                device_manager = await subscriber.async_get_device_manager()
                device = device_manager.devices[args.device_id]
                callback = DeviceWatcherCallback(device, args.output_type)
                device.add_event_callback(callback)
            else:
                subscriber.set_update_callback(SubscribeCallback(args.output_type))
            await subscriber.start_async()
            try:
                while True:
                    await asyncio.sleep(10)
            except KeyboardInterrupt:
                subscriber.stop_async()

        # All other commands require a device_id
        device = await api.async_get_device(args.device_id)

        if args.command == "get_device":
            PrintDevice(device, args.output_type)

        if args.command == "set_mode":
            mode = args.mode
            trait = device.traits[ThermostatModeTrait.NAME]
            if mode == "MANUAL_ECO":
                trait = device.traits[ThermostatEcoTrait.NAME]
            resp = await trait.set_mode(mode)
            print(await resp.text())

        if args.command == "set_heat":
            trait = device.traits[ThermostatTemperatureSetpointTrait.NAME]
            resp = await trait.set_heat(args.heat)
            print(await resp.text())

        if args.command == "set_cool":
            trait = device.traits[ThermostatTemperatureSetpointTrait.NAME]
            resp = await trait.set_cool(args.cool)
            print(await resp.text())

        if args.command == "set_range":
            trait = device.traits[ThermostatTemperatureSetpointTrait.NAME]
            resp = await trait.set_range(args.heat, args.cool)
            print(await resp.text())

        if args.command == "generate_rtsp_stream":
            trait = device.traits[CameraLiveStreamTrait.NAME]
            stream = await trait.generate_rtsp_stream()
            print(f"URL: {stream.rtsp_stream_url}")
            print(f"Stream Token: {stream.stream_token}")
            print(f"Expires At: {stream.expires_at}")


def main():
    args = parser.parse_args()
    if args.verbose:
        logging.basicConfig(level=logging.DEBUG)
    user_creds = CreateCreds(args)
    if args.creds_json_file:
        service_account.Credentials.from_service_account_file(
            args.creds_json_file
        ).with_scopes(scopes=SDM_SCOPES)
    else:
        service_creds = user_creds
    loop = asyncio.get_event_loop()
    loop.run_until_complete(RunTool(args, user_creds, service_creds))
    loop.close()


if __name__ == "__main__":
    main()



"""A device from the Smart Device Management API."""

import logging

# Import traits for registration
from typing import Callable

from . import camera_traits  # noqa: F401
from . import device_traits  # noqa: F401
from . import doorbell_traits  # noqa: F401
from . import thermostat_traits  # noqa: F401
from .auth import AbstractAuth
from .event import AsyncEventCallback, EventMessage, EventProcessingError
from .traits import BuildTraits, Command

_LOGGER = logging.getLogger(__name__)

DEVICE_NAME = "name"
DEVICE_TYPE = "type"
DEVICE_TRAITS = "traits"
DEVICE_PARENT_RELATIONS = "parentRelations"
PARENT = "parent"
DISPLAYNAME = "displayName"


class Device:
    """Class that represents a device object in the Google Nest SDM API."""

    def __init__(self, raw_data: dict, traits: dict):
        """Initialize a device."""
        self._raw_data = raw_data
        self._traits = traits
        self._relations = {}
        for relation in self._raw_data.get(DEVICE_PARENT_RELATIONS, []):
            if PARENT not in relation or DISPLAYNAME not in relation:
                continue
            self._relations[relation[PARENT]] = relation[DISPLAYNAME]
        self._callbacks = []

    @staticmethod
    def MakeDevice(raw_data: dict, auth: AbstractAuth):
        """Creates a device with the appropriate traits."""
        device_id = raw_data.get(DEVICE_NAME)
        cmd = Command(device_id, auth)
        traits = raw_data.get(DEVICE_TRAITS, {})
        traits_dict = BuildTraits(traits, cmd, raw_data.get(DEVICE_TYPE))
        return Device(raw_data, traits_dict)

    @property
    def name(self) -> str:
        """The resource name of the device such as 'enterprises/XYZ/devices/123'."""
        return self._raw_data[DEVICE_NAME]

    @property
    def type(self) -> str:
        """Type of device for display purposes.

        The device type should not be used to deduce or infer functionality of
        the actual device it is assigned to. Instead, use the returned traits for
        the device.
        """
        return self._raw_data[DEVICE_TYPE]

    @property
    def traits(self) -> dict:
        """Return a trait mixin or None."""
        return self._traits

    def _traits_data(self, trait) -> dict:
        """Return the raw dictionary for the specified trait."""
        traits_dict = self._raw_data.get(DEVICE_TRAITS, {})
        return traits_dict.get(trait, {})

    @property
    def parent_relations(self) -> dict:
        """"Assignee details of the device (e.g. room/structure)."""
        return self._relations

    @property
    def raw_data(self) -> str:
        """Return the raw data string."""
        return self._raw_data

    def add_event_callback(
        self, event_callback: AsyncEventCallback
    ) -> Callable[[], None]:
        """Register an AsyncEventCallback for udpates to this device.

        The return value is a callable that will unregister the callback.
        """
        self._callbacks.append(event_callback)

        def remove_callback():
            """Remove the event_callback."""
            self._callbacks.remove(event_callback)

        return remove_callback

    async def async_handle_event(self, event_message: EventMessage) -> None:
        """Process an event from the pubsub subscriber."""
        _LOGGER.debug(
            "Processing update %s @ %s", event_message.event_id, event_message.timestamp
        )
        if not event_message.resource_update_name:
            raise EventProcessingError("Event was not resource update event")
        if self.name != event_message.resource_update_name:
            raise EventProcessingError(
                f"Mismatch {self.name} != {event_message.resource_update_name}"
            )
        traits = event_message.resource_update_traits
        if traits:
            _LOGGER.debug("Trait update %s", traits.keys())
        events = event_message.resource_update_events
        if events:
            _LOGGER.debug("Event Update %s", events.keys())

        for (trait_name, trait) in traits.items():
            self._traits[trait_name] = trait

        for callback in self._callbacks:
            await callback.async_handle_event(event_message)
